;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(*
registers:
  BC -- IP
  SP -- data stack
  IX -- user area address

  for asm code i used disassembly done by Marcos Cruz (programandala.net)
  http://programandala.net/en.program.abersoft_forth.html

  original code by John Jones-Steel, 1983
*)

$use <zxasm>

;; use DTC instead of ITC?
true quan OPT-DTC?

;; relative BRANCHes?
;; non-relative branches are slightly faster
;; (and the code is smaller by almost 20 bytes)
false quan OPT-REL-BRANCH?

;; include debug ".HEXn" printers?
false quan OPT-DOT-HEX?

;; include "CLIT" and "BLIT" -- byte literals?
;; the code is very slightly smaller this way
true quan OPT-CLIT?

;; include various additional block words?
false quan OPT-MORE-BLOCK-WORDS?

;; include editor code?
false quan OPT-EDITOR-CODE?

;; include "eratos test"?
false quan OPT-ETEST?


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; argument parsing

$use <getarg>

0 quan output-file-name ;; dynalloced

|: set-output-fname  ( addr count )
  output-file-name ?error" too many output files" string:$new output-file-name:! ;

getarg:arg: --dtc OPT-DTC?:!t ;
OPT-DTC? [IFNOT] getarg:help: ." use Direct Threaded Code" ; [ENDIF]
getarg:arg: --itc OPT-DTC?:!f ;
OPT-DTC? [IF] getarg:help: ." use Indirect Threaded Code" ; [ENDIF]

getarg:arg: --rel-branches OPT-REL-BRANCH?:!t ;
OPT-REL-BRANCH? [IFNOT] getarg:help: ." use relative branch offsets" ; [ENDIF]
getarg:arg: --abs-branches OPT-REL-BRANCH?:!f ;
OPT-REL-BRANCH? [IF] getarg:help: ." use absolute branch offsets" ; [ENDIF]

getarg:arg: --clit OPT-CLIT?:!t ;
OPT-CLIT? [IFNOT] getarg:help: ." include CLIT and BLIT byte-sized literal primitives" ; [ENDIF]
getarg:arg: --no-clit OPT-CLIT?:!f ;
OPT-CLIT? [IF] getarg:help: ." do not include CLIT and BLIT byte-sized literal primitives" ; [ENDIF]

getarg:arg: --debug-dot-hex OPT-DOT-HEX?:!t ;
OPT-DOT-HEX? [IFNOT] getarg:help: ." include .HEX2 and .HEX4 words" ; [ENDIF]
getarg:arg: --no-debug-dot-hex OPT-DOT-HEX?:!f ;
OPT-DOT-HEX? [IF] getarg:help: ." do not include .HEX2 and .HEX4 words" ; [ENDIF]

getarg:arg: --extra-bwords OPT-MORE-BLOCK-WORDS?:!t ;
OPT-MORE-BLOCK-WORDS? [IFNOT] getarg:help: ." include some extra block words" ; [ENDIF]
getarg:arg: --no-extra-bwords OPT-MORE-BLOCK-WORDS?:!f ;
OPT-MORE-BLOCK-WORDS? [IF] getarg:help: ." do not include some extra block words" ; [ENDIF]

getarg:arg: --editor OPT-EDITOR-CODE?:!t ;
OPT-MORE-BLOCK-WORDS? [IFNOT] getarg:help: ." include EDITOR words" ; [ENDIF]
getarg:arg: --no-editor OPT-EDITOR-CODE?:!f ;
OPT-MORE-BLOCK-WORDS? [IF] getarg:help: ." do not include EDITOR words" ; [ENDIF]

getarg:arg: --output  ( addr count )  set-output-fname ;
getarg:last-need-arg
getarg:help: ." output tape file (default is \'tap/aforth.tap\')" ;
getarg:last-alias: -o

;; hidden options
getarg:arg: --etest OPT-ETEST?:!t ;
getarg:arg: --no-etest OPT-ETEST?:!f ;

getarg:usage:  ." urforth aber-main.f [args]\nargs:\n" ;
getarg:parse-args

output-file-name [IFNOT]
" tap/aforth.tap" set-output-fname
[ENDIF]

: on/off.  ( flag )  ?< ." en" || ." dis" >? ." abled.\n" ;

: opt.  ( addr count flag )  endcr >r type ." : " r> on/off. ;

: options.
  endcr
  ." Threaded Code Type: " OPT-DTC? ?< ." DTC\n" || ." ITC\n" >?
  ." Branch Offsets: " OPT-REL-BRANCH? ?< ." relative\n" || ." absolute\n" >?
   " CLIT/BLIT" OPT-CLIT? opt.
   " .HEX2/.HEX4" OPT-DOT-HEX? opt.
   " Extra Block Words" OPT-MORE-BLOCK-WORDS? opt.
   " Editor Words" OPT-EDITOR-CODE? opt. ;
options.


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; main code

;; define some constants
module TCOM
;; base address for the new system
$6000 quan zx-base-addr

$CB40 constant zx-s0
$CBE0 constant zx-r0

zx-r0 constant zx-first-buffer
$D000 constant zx-buffer-limit
zx-buffer-limit constant zx-ramdisc^
$2C00 constant zx-#ramdisc

zx-buffer-limit constant zx-buffer-lo
$FBFF constant zx-buffer-hi

zx-s0 constant zx-query-tib
end-module


$include "tcom-10-base.f"

$include "tcom-20-creatori.f"

$include "figsrc/zx-10-startup.f"
tcom:create-first-zx-word

tcom:zx-vocabulary EDITOR tcom:zx-immediate
tcom:zx-latest-vocid tcom:zx-editor-vocid:!

$include "figsrc/zx-20-prims.f"
$include "figsrc/zx-22-prims-io.f"
$include "figsrc/zx-24-prims-muldiv.f"
$include "figsrc/zx-26-prims-math.f"
$include "figsrc/zx-28-prims-mem.f"

$include "tcom-30-if-begin.f"

$include "figsrc/zx-30-low-level.f"
$include "figsrc/zx-32-blocks-and-more.f"
$include "figsrc/zx-34-if-then.f"
$include "figsrc/zx-40-creatori.f"
$include "figsrc/zx-50-interpret.f"
$include "figsrc/zx-60-other.f"

$include "tcom-80-see.f"

OPT-EDITOR-CODE? [IF]
$include "figsrc/zx-70-editor.f"
[ENDIF]
OPT-ETEST? [IF]
$include "figsrc/zx-96-etest.f"
[ENDIF]

\ tcom:zx-see VOCABULARY
\ tcom:zx-see definitions
\ tcom:zx-see [
\ tcom:zx-see abort
\ tcom:zx-see interpret
\ tcom:zx-see quit
\ tcom:zx-see expect
\ tcom:zx-see key
\ tcom:zx-see type

tcom:check-forwards

tcom:zx-dp^ tcom:zx-w@ tcom:zx-fence^ tcom:zx-w!

;; DP, FENCE, VOCLINK
tcom:zx-dp^ tcom:zx-w@ tcom:zx-dpx-init^ tcom:zx-w!
tcom:zx-dp^ 2 + tcom:zx-w@ tcom:zx-dpx-init^ 2 + tcom:zx-w!
tcom:zx-dp^ 4 + tcom:zx-w@ tcom:zx-dpx-init^ 4 + tcom:zx-w!
;; FORTH-LATEST
tcom:zx-forth-vocid tcom:zx-w@ tcom:zx-dpx-init^ 6 + tcom:zx-w!

\ tcom:zx-see COLD

tcom:zx-here tcom:zx-base-addr -
." MAIN SYSTEM size: " ., ." bytes.\n"

zxa:finalise
output-file-name count
endcr ." SAVING TAPE: " 2dup type cr
zxa:writers:tap-save

